#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2012 - 2013
# Matías Herranz <matiasherranz@gmail.com>
# Joaquín Tita <joaquintita@gmail.com>
# https://github.com/PyRadar/pyradar
# 
# 2020: Erdogan Taskesen: <erdogant@gmail.com> Converted to python3.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <http://www.gnu.org/licenses/>.


from math import exp
import numpy as np
import logging

logger = logging.getLogger(__name__)

K_DEFAULT = 1.0
CU_DEFAULT = 0.523
CMAX_DEFAULT = 1.73

def weighting(pix_value, window, k=K_DEFAULT, cu=CU_DEFAULT, cmax=CMAX_DEFAULT):
    """
    Computes the weighthing function for Lee filter using cu as the noise
    coefficient.
    """

    # cu is the noise variation coefficient

    # ci is the variation coefficient in the window
    window_mean = window.mean()
    window_std = window.std()
    try:
        ci = window_std / window_mean
    except:
        ci = 0

    if ci <= cu:  # use the mean value
        w_t = 1.0
    elif cu < ci < cmax:  # use the filter
        w_t = exp((-k * (ci - cu)) / (cmax - ci))
    elif ci >= cmax:  # preserve the original value
        w_t = 0.0
    else: # use the mean value
        w_t = 1.0

    return w_t


def assert_parameters(win_size, k, cu, cmax):
    """Asserts parameters in range.
    
    Parameters
    ----------
    win_size : int
        Window size parameter.
    k : float
        k parameter in range [0:10].
    cu : float
        cu parameter, must be positive.
    cmax : float
        cmax parameter, must be positive and greater equal than cu.
    """
    if 0 > k > 10: raise Exception("k parameter out of range 0<= k <= 10, submitted %s" %(k))
    if cu < 0: raise Exception( "cu can't be negative")
    if cmax < 0 and cmax < cu: raise Exception("cmax must be positive and greater equal to cu: %s" %(cu))
    if win_size < 3: raise Exception('ERROR: win size must be at least 3')


def lee_enhanced_filter(img, win_size=3, k=K_DEFAULT, cu=CU_DEFAULT, cmax=CMAX_DEFAULT):
    """Lee enhanced filter.

    Enhanced Lee Filter Description
    -------------------------------
    Apply Enhanced Lee filter to a numpy matrix containing the image, with a window of win_size x win_size.

    Parameters
    ----------
    img : array-like
        Input image.
    win_size : int, int (default: 3)
        Window size.
    cu : float (default: 0.25)
        cu factor.
    k : float, (default: 1.0)
        Kvalue.
    cmax : float, (default: 1.73)
        cmax value.

    Returns
    -------
    img_filtered : array-like
        Filtered image.

    Examples
    --------
    >>> import findpeaks
    >>> import matplotlib.pyplot as plt
    >>> img = findpeaks.import_example('2dpeaks_image')
    >>> # Resize
    >>> img = findpeaks.stats.resize(img, size=(300,300))
    >>> # Make grey image
    >>> img = findpeaks.stats.togray(img)
    >>> # Scale between [0-255]
    >>> img = findpeaks.stats.scale(img)
    >>> # Filter
    >>> img_filtered = findpeaks.stats.lee_enhanced_filter(img.copy(), win_size=15)
    >>>
    >>> plt.figure()
    >>> fig, axs = plt.subplots(1,2)
    >>> axs[0].imshow(img, cmap='gray'); axs[0].set_title('Input')
    >>> axs[1].imshow(img_filtered, cmap='gray'); axs[1].set_title('Lee enhanced filter')

    """
    if win_size < 3: raise Exception('[findpeaks] >ERROR: win size must be at least 3')
    if len(img.shape) > 2: raise Exception('[findpeaks] >ERROR: Image should be 2D. Hint: set the parameter: togray=True')
    if ((win_size % 2) == 0): print('[findpeaks] >It is highly recommended to user odd window sizes. You provided %s, an even number.' % (win_size))
    assert_parameters(win_size, k, cu, cmax)

    # we process the entire img as float64 to avoid type overflow error
    img = np.float64(img)
    img_filtered = np.zeros_like(img)
    N, M = img.shape
    win_offset = int(win_size / 2)

    for i in np.arange(0, N):
        xleft = i - win_offset
        xright = i + win_offset

        if xleft < 0:
            xleft = 0
        if xright >= N:
            xright = N

        for j in np.arange(0, M):
            yup = j - win_offset
            ydown = j + win_offset

            if yup < 0:
                yup = 0
            if ydown >= M:
                ydown = M

            pix_value = img[i, j]
            window = img[xleft:xright, yup:ydown]
            w_t = weighting(pix_value, window, k, cu, cmax)
            window_mean = window.mean()

            new_pix_value = (window_mean * w_t) + (pix_value * (1.0 - w_t))

            if new_pix_value < 0.0: raise Exception("ERROR: lee_filter(), pixel filtered can't be negative")
            if (new_pix_value is None) or np.isnan(new_pix_value):
                new_pix_value = 0

            img_filtered[i, j] = round(new_pix_value)

    return img_filtered



from findpeaks import findpeaks
import numpy as np
import unittest
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')  # Use non-interactive backend for tests

# Configure matplotlib to allow more open figures before warning
matplotlib.rcParams['figure.max_open_warning'] = 100


class TestFINDPEAKS_BLOG(unittest.TestCase):

    def tearDown(self):
        """Clean up matplotlib figures after each test to prevent memory issues."""
        try:
            plt.close('all')
            # Also close any remaining figures that might not be caught by 'all'
            for fig_num in plt.get_fignums():
                plt.figure(fig_num).close()
        except Exception:
            # Ignore any errors during cleanup to avoid masking test failures
            pass

    def example1(self):
        # Import library
        from findpeaks import findpeaks
        # Initialize
        fp = findpeaks(method='topology')
        # Example 1d-vector
        X = fp.import_example('1dpeaks')
        
        # Plot
        plt.figure(figsize=(15, 8), dpi=100)
        plt.plot(X)
        plt.grid(True)
        plt.xlabel('Time')
        plt.ylabel('Value')

        # Fit topology method on the 1D vector
        results = fp.fit(X)
        # Plot the results
        fp.plot_persistence(figsize=(25,8), fontsize_ax2=18, s=100)
    
    def example2(self):
        # Import findpeaks
        from findpeaks import findpeaks
        # URL to image
        path = r'https://erdogant.github.io/datasets/images/complex_peaks.png'
        # Set findpeaks with its parameters
        fp = findpeaks(method='topology', whitelist='peak', limit=5, denoise='lee_sigma', params={'window': 5})
        # Read image from url
        X = fp.imread(path)
        # Detect peaks
        results = fp.fit(X)
        # Show persistence plot
        fp.plot_persistence()
        # Show plot
        fp.plot(figsize=(25, 14), text=True, marker='x', color='#ff0000', figure_order='vertical', fontsize=10)
        # Mesh plot
        fp.plot_mesh(view=(40, 225), cmap='hot')
        # Results in dataframe
        result_df = results['persistence']
        
        """
              x    y  birth_level  death_level       score  peak valley
        0   433   95       255.00     0.000000  255.000000  True  False
        1   310   95       255.00     8.920000  246.080000  True  False
        2    62   95       255.00     8.960000  246.040000  True  False
        3   186   95       255.00     9.680000  245.320000  True  False
        4   457  100        85.00    33.600000   51.400000  True  False
        5    39  100        85.00    33.640000   51.360000  True  False
        6   163  100        85.00    34.040000   50.960000  True  False
        7   334  100        85.00    34.160000   50.840000  True  False
        8   210  100        85.00    34.240000   50.760000  True  False
        9   286  100        85.00    34.320000   50.680000  True  False
        ...
        ...
        46   13   50        15.16     9.833333    5.326667  True  False
        """
    
    def example3(self):
        # Initialize
        fp = findpeaks(method='mask')
        # Example 2d image
        X = fp.import_example('2dpeaks')
        # Plot RAW input image
        plt.imshow(X)
        # Fit using mask method
        results = fp.fit(X)
        # Plot the pre-processing steps
        fp.plot_preprocessing()
        # The output contains multiple variables
        # Plot detected peaks
        fp.plot(figure_order='horizontal', fontsize=14)
        # Create mesh plot from 2D image.
        fp.plot_mesh()
        # Rotate to make a top view
        fp.plot_mesh(view=(90,0))
    
    def examples3(self):
        # Import libraries
        import numpy as np
        from findpeaks import findpeaks
        
        # Create example data set
        i = 10000
        xs = np.linspace(0,3.7*np.pi,i)
        X = (0.3*np.sin(xs) + np.sin(1.3 * xs) + 0.9 * np.sin(4.2 * xs) + 0.06 * np.random.randn(i))
        # Initialize
        fp = findpeaks(method='peakdetect', lookahead=200, interpolate=None)
        # Fit peakdetect method
        results = fp.fit(X)
        # Plot
        fp.plot()
    
    def examples4(self):
        # Import library
        from findpeaks import findpeaks
        
        # Initialize findpeaks with cearus method.
        # The default setting is that it only return peaks-vallyes with at least 5% difference.
        fp = findpeaks(method='caerus', params={'minperc':5, 'window':50})
        # Import example data
        X = fp.import_example('facebook')
        # Fit
        results = fp.fit(X)
        # Make the plot
        fp.plot()
    
    def examples5(self):
        # Import libraries
        import numpy as np
        from findpeaks import findpeaks
        
        # Create example data set
        i = 10000
        xs = np.linspace(0,3.7*np.pi,i)
        X = (0.3*np.sin(xs) + np.sin(1.3 * xs) + 0.9 * np.sin(4.2 * xs) + 0.06 * np.random.randn(i))
        # Initialize
        fp = findpeaks(method='peakdetect', lookahead=200, interpolate=None)
        # Fit peakdetect method
        results = fp.fit(X)
        # Plot
        fp.plot()
    
    def examples6(self):
        # Import library
        from findpeaks import findpeaks
        
        # Initialize findpeaks with cearus method.
        # The default setting is that it only return peaks-vallyes with at least 5% difference.
        fp = findpeaks(method='caerus', params={'minperc':5, 'window':50})
        # Import example data
        X = fp.import_example('facebook')
        # Fit
        results = fp.fit(X)
        # Make the plot
        fp.plot()
    
    def examples7(self):
        # Import library
        from findpeaks import findpeaks
        # Initializatie
        fp = findpeaks(scale=None, denoise=None, togray=True, imsize=(300, 300))
        # Import image example
        img = fp.import_example('2dpeaks_image')
        # Fit
        fp.fit(img)
        # Tens of thousands of peaks are detected at this point. Better to put text=False
        fp.plot(figure_order='horizontal', text=False, alpha=0.05)
        fp.plot_mesh()
    
        # Import library
        # Initializatie
        fp = findpeaks(method='topology',
                       togray=True,
                       imsize=(300, 300),
                       scale=True,
                       denoise='fastnl',
                       params={'window': 31})
        
        # Import image example
        img = fp.import_example('2dpeaks_image')
        # Fit
        fp.fit(img)
        # Plot
        fp.plot_preprocessing()
        
        # Plot the top 15 peaks that are detected and examine the scores
        fp.results['persistence'][0:5]
        
        # Take the minimum score for the top peaks off the diagonal.
        limit = fp.results['persistence'][0:5]['score'].min()
        # Plot
        fp.plot(limit=limit, figure_order='horizontal')
        # Mesh plot
        fp.plot_mesh()

    def example9(self):
        # Import library
        from findpeaks import findpeaks
        # Initialize findpeaks with cearus method.
        # The default setting is that it only return peaks-vallyes with at least 5% difference. We can change this using params
        # fp = findpeaks(method='caerus',  params={'minperc': 10, 'window': 50})
        fp = findpeaks(method='caerus')
        # Import example data
        X = fp.import_example('facebook')
        # Fit
        results = fp.fit(X)
        # Make the plot
        fp.plot()
        

class TestFINDPEAKS(unittest.TestCase):
    
    def tearDown(self):
        """Clean up matplotlib figures after each test to prevent memory issues."""
        try:
            plt.close('all')
            # Also close any remaining figures that might not be caught by 'all'
            for fig_num in plt.get_fignums():
                plt.figure(fig_num).close()
        except Exception:
            # Ignore any errors during cleanup to avoid masking test failures
            pass

    def test_fit(self):
        # CHECK OUTPUT METHOD TOPOLOGY
        import numpy as np
        from findpeaks import findpeaks

        fp = findpeaks(method="topology", whitelist=['peak'])
        X = fp.import_example('2dpeaks')
        results = fp.fit(X)
        assert fp.type == 'peaks2d'
        assert [*results.keys()] == ['Xraw', 'Xproc', 'Xdetect', 'Xranked', 'persistence', 'groups0']
        assert [*fp.args] == ['limit', 'scale', 'denoise', 'togray', 'imsize', 'figsize', 'type']
        assert results['Xraw'].shape == results['Xdetect'].shape
        assert results['Xproc'].shape == results['Xdetect'].shape

        fp.plot(figsize=(25, 15), figure_order='horizontal')

        # CHECK RESULTS METHOD TOPOLOGY
        assert len(results['Xdetect'][results['Xdetect'] != 0]) > 18
        assert len(results['Xranked'][results['Xranked'] != 0]) > 18

    def test_topology_limit(self):
        from findpeaks import findpeaks
        # CHECK RESULTS METHOD with LIMIT functionality
        fp = findpeaks(method="topology", whitelist=['peak'], limit=0)
        X = fp.import_example('2dpeaks')
        results = fp.fit(X)
        fp.plot(figsize=(25, 15), figure_order='horizontal')
        assert len(results['Xdetect'][results['Xdetect'] != 0]) > 18
        assert len(results['Xranked'][results['Xranked'] != 0]) > 18

    def test_mask_2dpeaks(self):
        import numpy as np
        from findpeaks import findpeaks
        # CHECK OUTPUT METHOD MASK
        fp = findpeaks(method="mask")
        X = fp.import_example('2dpeaks')
        results = fp.fit(X)
        assert fp.type == 'peaks2d'
        assert [*results.keys()] == ['Xraw', 'Xproc', 'Xdetect', 'Xranked']
        assert [*fp.args] == ['limit', 'scale', 'denoise', 'togray', 'imsize', 'figsize', 'type']
        fp.plot(figsize=(25, 15), figure_order='horizontal')

        # CHECK RESULTS METHOD TOPOLOGY
        assert np.sum(results['Xdetect']) == 20
        assert results['Xraw'].shape == results['Xdetect'].shape
        assert results['Xproc'].shape == results['Xdetect'].shape

    def test_whitelist(self):
        # CHECK WHITELIST
        import numpy as np
        from scipy.ndimage import gaussian_filter
        from findpeaks import findpeaks
        rng = np.random.default_rng(42)
        x = rng.normal(size=(50, 50))
        x = gaussian_filter(x, sigma=10.)
        # peak and valley
        fp = findpeaks(method="topology", whitelist=['peak', 'valley'], denoise=None)
        results = fp.fit(x)

        fp.plot(figsize=(25, 15), figure_order='horizontal')
        fp.plot_persistence()
        fp.plot_mesh()

        Iloc = results['persistence']['score'] > 1
        assert results['persistence']['peak'][Iloc].sum() >0
        assert results['persistence']['valley'][Iloc].sum() >0

        # peaks
        fp = findpeaks(method="topology", whitelist='peak', denoise=None)
        fp.plot()
        results = fp.fit(x)
        Iloc = results['persistence']['score'] > 1
        assert results['persistence']['peak'][Iloc].shape[0] == results['persistence']['peak'][Iloc].sum()

        fp = findpeaks(method="topology", whitelist='valley', denoise=None)
        results = fp.fit(x)
        Iloc = results['persistence']['score'] > 1
        assert results['persistence']['valley'].shape[0] == results['persistence']['valley'].sum()

    def test_topology(self):
        # CHECK OUTPUT METHOD TOPOLOGY
        fp = findpeaks(method="topology")
        X = fp.import_example('1dpeaks')
        results = fp.fit(X)
        assert fp.type == 'peaks1d'
        assert [*results.keys()] == ['persistence', 'Xdetect', 'Xranked', 'groups0', 'df']
        assert [*fp.args] == ['method', 'params', 'lookahead', 'interpolate', 'figsize', 'type']
        assert len(X) == len(results['Xdetect'])
        assert len(X) == len(results['Xranked'])
        assert len(X) == results['df'].shape[0]
        assert np.all(np.isin(results['df'].columns, ['x', 'y', 'labx', 'rank', 'score', 'valley', 'peak']))
        assert np.all(np.isin(results['persistence'].columns, ['x', 'y', 'birth_level', 'death_level', 'score']))

        # CHECK RESULTS METHOD TOPOLOGY
        # Let op, soms gaat deze ook op 6 vanwege een stochastic components
        # assert results['persistence'].shape[0] == 7

        # CHECK RESULTS METHOD with LIMIT functionality
        X = fp.import_example('1dpeaks')
        fp = findpeaks(method="topology", limit=0.02)
        results = fp.fit(X)
        assert len(results['Xdetect'][results['Xdetect'] != 0]) == len(results['Xranked'][results['Xranked'] != 0])

        # CHECK OUTPUT METHOD PEAKDETECT
        # fp = findpeaks(method="peakdetect", lookahead=1, verbose=3, height=0)
        fp = findpeaks(method="peakdetect", lookahead=1)
        X = fp.import_example('1dpeaks')
        results = fp.fit(X)
        assert fp.type == 'peaks1d'
        assert [*results.keys()] == ['df']
        assert [*fp.args] == ['method', 'params', 'lookahead', 'interpolate', 'figsize', 'type']
        assert len(X) == results['df'].shape[0]
        assert np.all(np.isin(results['df'].columns, ['x', 'y', 'labx', 'valley', 'peak', 'rank', 'score']))

        # CHECK RESULTS METHOD TOPOLOGY
        assert results['df']['peak'].sum() == 2
        assert results['df']['valley'].sum() == 4
        fp.plot()

        # Run over all combinations and make sure no errors are made
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        methods = ['topology', 'peakdetect', None]
        interpolates = [None, 1, 10, 1000]
        lookaheads = [None, 0, 1, 10, 100]
        for method in methods:
            for interpolate in interpolates:
                for lookahead in lookaheads:
                    fp = findpeaks(method=method, lookahead=lookahead, interpolate=interpolate)
                    assert fp.fit(X)
    
    def test_method_combinations(self):
        """Test all method combinations with various parameters"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        
        # Test different methods
        methods = ['peakdetect', 'caerus', None]
        for method in methods:
            fp = findpeaks(method=method)
            results = fp.fit(X)
            assert results is not None
    
    def test_verbose_levels(self):
        """Test different verbose levels"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        verbose_levels = ['info', 'warning', 'error', 'debug', 0, 1, 2, 3]
        
        for verbose in verbose_levels:
            fp = findpeaks(method='peakdetect', verbose=verbose)
            results = fp.fit(X)
            assert results is not None
    
    def test_figsize_parameters(self):
        """Test different figsize parameters"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        figsize_options = [(10, 5), (20, 10), (5, 15), None]
        
        for figsize in figsize_options:
            fp = findpeaks(method='peakdetect', figsize=figsize)
            results = fp.fit(X)
            assert results is not None
    
    def test_whitelist_combinations(self):
        """Test different whitelist combinations"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        whitelist_options = ['peak', 'valley', ['peak', 'valley'], None]
        
        for whitelist in whitelist_options:
            fp = findpeaks(method='topology', whitelist=whitelist)
            results = fp.fit(X)
            assert results is not None
    
    def test_limit_parameters(self):
        """Test different limit parameters"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        limit_options = [0, 0.1, 0.5, 1.0, 10, None]
        
        for limit in limit_options:
            fp = findpeaks(method='topology', limit=limit)
            results = fp.fit(X)
            assert results is not None
    
    def detection_1d(self):
        from findpeaks import findpeaks
        X = [1,1,1.1,1,0.9,1,1,1.1,1,0.9,1,1.1,1,1,0.9,1,1,1.1,1,1,1,1,1.1,0.9,1,1.1,1,1,0.9,1,1.1,1,1,1.1,1,0.8,0.9,1,1.2,0.9,1,1,1.1,1.2,1,1.5,1,3,2,5,3,2,1,1,1,0.9,1,1,3,2.6,4,3,3.2,2,1,1,0.8,4,4,2,2.5,1,1,1]
        
        fp = findpeaks(method='peakdetect', whitelist=['peak', 'valley'], lookahead=2, verbose='info')
        results = fp.fit(X)
        fp.plot()
        assert results['df'].shape == (74, 5)
        assert results['df']['valley'].sum()== 11
        assert results['df']['peak'].sum()== 9    

    def test_interpolation_parameters(self):
        """Test different interpolation parameters"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        interpolate_options = [1, 2, 3, 5, 10, 100, None]
        
        for interpolate in interpolate_options:
            fp = findpeaks(method='peakdetect', interpolate=interpolate)
            results = fp.fit(X)
            assert results is not None
    
    def test_lookahead_parameters(self):
        """Test different lookahead parameters"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        lookahead_options = [0, 1, 2, 5, 10, 50, 100, 200, None]
        
        for lookahead in lookahead_options:
            fp = findpeaks(method='peakdetect', lookahead=lookahead)
            results = fp.fit(X)
            assert results is not None
    
    def test_delta_parameters(self):
        """Test different delta parameters for peakdetect"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        delta_options = [0, 0.1, 0.5, 1.0, 5.0, 10.0]
        
        for delta in delta_options:
            fp = findpeaks(method='peakdetect', params={'delta': delta})
            results = fp.fit(X)
            assert results is not None
    
    def test_window_parameters(self):
        """Test different window parameters for various filters"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        window_options = [3, 5, 7, 9, 11, 15, 21, None]
        
        for window in window_options:
            fp = findpeaks(method='peakdetect', params={'window': window})
            results = fp.fit(X)
            assert results is not None

    def detection_2d(self):
        from findpeaks import findpeaks
        import matplotlib.pyplot as plt
        
        # Initialize
        fp = findpeaks(method='topology', imsize=(150, 150), scale=True, togray=True, denoise='lee_sigma', params={'window': 17})
        # Import example image
        img = fp.import_example('2dpeaks_image')
        # Denoising and detecting peaks
        results = fp.fit(img)
        # Create mesh plot
        fp.plot_mesh()
        # Create denoised plot
        fp.plot(limit=80, figure_order='horizontal', cmap=plt.cm.hot_r)
        assert sum(results['persistence']['score']>80)==3
    
    def test_2d_parameters(self):
        """Test different 2D parameters"""
        fp = findpeaks()
        X = fp.import_example('2dpeaks')
        
        # Test different imsize parameters
        imsize_options = [(100, 100), (200, 200), (300, 300), None]
        for imsize in imsize_options:
            fp = findpeaks(method='topology', imsize=imsize)
            results = fp.fit(X)
            assert results is not None
        
        # Test scale parameter
        for scale in [True, False]:
            fp = findpeaks(method='topology', scale=scale)
            results = fp.fit(X)
            assert results is not None
        
        # Test togray parameter
        for togray in [True, False]:
            fp = findpeaks(method='topology', togray=togray)
            results = fp.fit(X)
            assert results is not None
    
    def test_scale_false_with_opencv_denoising(self):
        """Test that scale=False works with OpenCV denoising methods"""
        fp = findpeaks()
        X = fp.import_example('2dpeaks')
        
        # Test with fastnl denoising (OpenCV method)
        fp = findpeaks(method='topology', scale=False, denoise='fastnl')
        results = fp.fit(X)
        assert results is not None
        
        # Test with bilateral denoising (OpenCV method)
        fp = findpeaks(method='topology', scale=False, denoise='bilateral')
        results = fp.fit(X)
        assert results is not None
        
        # Test with non-OpenCV denoising methods
        non_opencv_methods = ['lee', 'lee_enhanced', 'lee_sigma', 'kuan', 'frost', 'median', 'mean']
        for method in non_opencv_methods:
            fp = findpeaks(method='topology', scale=False, denoise=method)
            results = fp.fit(X)
            assert results is not None
    
    def test_denoising_methods(self):
        """Test all denoising methods"""
        fp = findpeaks()
        X = fp.import_example('2dpeaks')
        
        denoise_methods = ['fastnl', 'bilateral', 'lee', 'lee_enhanced', 'lee_sigma', 
                          'kuan', 'frost', 'median', 'mean', None]
        
        for method in denoise_methods:
            fp = findpeaks(method='topology', denoise=method)
            results = fp.fit(X)
            assert results is not None
    
    def test_denoising_parameters(self):
        """Test different denoising parameters"""
        fp = findpeaks()
        X = fp.import_example('2dpeaks')
        
        # Test different window sizes for denoising
        window_sizes = [3, 5, 7, 9, 11, 15, 21, None]
        for window in window_sizes:
            fp = findpeaks(method='topology', denoise='lee', params={'window': window})
            results = fp.fit(X)
            assert results is not None
        
        # Test different cu values for lee filter
        cu_values = [0.1, 0.25, 0.5, 0.75, 1.0, None]
        for cu in cu_values:
            fp = findpeaks(method='topology', denoise='lee', params={'cu': cu})
            results = fp.fit(X)
            assert results is not None
    
    def test_frost_filter_parameters(self):
        """Test frost filter specific parameters"""
        fp = findpeaks()
        X = fp.import_example('2dpeaks')
        
        # Test different damping factors
        damping_factors = [1.0, 2.0, 3.0, 5.0]
        for damping in damping_factors:
            fp = findpeaks(method='topology', denoise='frost', 
                          params={'damping_factor': damping, 'win_size': 15})
            results = fp.fit(X)
            assert results is not None
    
    def test_lee_enhanced_parameters(self):
        """Test lee enhanced filter specific parameters"""
        fp = findpeaks()
        X = fp.import_example('2dpeaks')
        
        # Test different k values
        k_values = [0.5, 1.0, 1.5, 2.0]
        for k in k_values:
            fp = findpeaks(method='topology', denoise='lee_enhanced', 
                          params={'k': k, 'cu': 0.523, 'cmax': 1.73, 'win_size': 15})
            results = fp.fit(X)
            assert results is not None
    
    def detection_2d_url(self):
        from findpeaks import findpeaks
        path = r'https://erdogant.github.io/datasets/images/complex_peaks.png'
        fp = findpeaks(method='topology', whitelist='peak', limit=5, denoise='lee_sigma', params={'window': 5})
        X = fp.imread(path)
        results = fp.fit(X)
        
        fp.plot_persistence()
        fp.plot(figsize=(25, 14), text=False, marker='x', color='#ff0000', figure_order='vertical', fontsize=14)
        # fp.plot_mesh(view=(40, 180))
        # fp.plot_mesh(view=(90, 0))
        assert results['persistence'].shape == (47, 7)

    def test_edge_cases(self):
        """Test edge cases and boundary conditions"""
        # Test with very small data
        X_small = [1, 2, 1]
        fp = findpeaks(method='peakdetect')
        results = fp.fit(X_small)
        assert results is not None
        
        # Test with constant data
        X_constant = [1, 1, 1, 1, 1]
        results = fp.fit(X_constant)
        assert results is not None
        
        # Test with single value
        X_single = [1]
        results = fp.fit(X_single)
        assert results is not None
        
        # Test with empty list
        X_empty = []
        try:
            results = fp.fit(X_empty)
            # Should either work or raise an exception
        except:
            pass  # Expected for empty data
    
    def test_data_types(self):
        """Test different data types"""
        # Test with numpy array
        import numpy as np
        X_np = np.array([10, 11, 9, 23, 21, 11, 45, 20, 11, 12])
        fp = findpeaks(method='peakdetect')
        results = fp.fit(X_np)
        assert results is not None
        
        # Test with list of floats
        X_float = [10.0, 11.0, 9.0, 23.0, 21.0, 11.0, 45.0, 20.0, 11.0, 12.0]
        results = fp.fit(X_float)
        assert results is not None
        
        # Test with mixed types
        X_mixed = [10, 11.0, 9, 23.0, 21, 11.0, 45, 20.0, 11, 12.0]
        results = fp.fit(X_mixed)
        assert results is not None
    
    def test_parameter_combinations(self):
        """Test various parameter combinations"""
        X = [9,60,377,985,1153,672,501,1068,1110,574,135,23,3,47,252,812,1182,741,263,33]
        
        # Test topology with different combinations
        combinations = [
            {'method': 'topology', 'limit': 0.1, 'whitelist': 'peak'},
            {'method': 'topology', 'limit': 0.5, 'whitelist': ['peak', 'valley']},
            {'method': 'peakdetect', 'lookahead': 1, 'interpolate': 3},
            {'method': 'peakdetect', 'lookahead': 2, 'params': {'delta': 0.5}},
        ]
        
        for combo in combinations:
            fp = findpeaks(**combo)
            results = fp.fit(X)
            assert results is not None
    
    def test_import_examples(self):
        """Test all available example datasets"""
        fp = findpeaks()
        
        # Test all available examples
        examples = ['1dpeaks', '2dpeaks', '2dpeaks_image']
        for example in examples:
            try:
                X = fp.import_example(example)
                assert X is not None
                # Test that we can fit on the imported data
                results = fp.fit(X)
                assert results is not None
            except Exception as e:
                # Some examples might not be available in all environments
                print(f"Example {example} not available: {e}")
    
    def test_denoising(self):

        # DENOISING METHODS TEST
        from findpeaks import findpeaks
        fp = findpeaks()
        img = fp.import_example('2dpeaks_image')
        import findpeaks

        # filters parameters
        # window size
        winsize = 15
        # damping factor for frost
        k_value1 = 2.0
        # damping factor for lee enhanced
        k_value2 = 1.0
        # coefficient of variation of noise
        cu_value = 0.25
        # coefficient of variation for lee enhanced of noise
        cu_lee_enhanced = 0.523
        # max coefficient of variation for lee enhanced
        cmax_value = 1.73

        # Some pre-processing
        # Resize
        img = findpeaks.stats.resize(img, size=(300, 300))
        # Make grey image
        img = findpeaks.stats.togray(img)
        # Scale between [0-255]
        img = findpeaks.stats.scale(img)

        # Denoising
        # fastnl
        img_fastnl = findpeaks.stats.denoise(img.copy(), method='fastnl', window=winsize)
        # bilateral
        img_bilateral = findpeaks.stats.denoise(img.copy(), method='bilateral', window=winsize)
        # frost filter
        image_frost = findpeaks.stats.frost_filter(img.copy(), damping_factor=k_value1, win_size=winsize)
        # kuan filter
        image_kuan = findpeaks.stats.kuan_filter(img.copy(), win_size=winsize, cu=cu_value)
        # lee filter
        image_lee = findpeaks.stats.lee_filter(img.copy(), win_size=winsize, cu=cu_value)
        # lee enhanced filter
        image_lee_enhanced = findpeaks.stats.lee_enhanced_filter(img.copy(), win_size=winsize, k=k_value2, cu=cu_lee_enhanced, cmax=cmax_value)
        # lee sigma filter
        image_lee_sigma = findpeaks.stats.lee_sigma_filter(img.copy())
        # mean filter
        image_mean = findpeaks.stats.mean_filter(img.copy(), win_size=winsize)
        # median filter
        image_median = findpeaks.stats.median_filter(img.copy(), win_size=winsize)

        # Loop throughout many combinations of parameter settings
        from findpeaks import findpeaks
        methods = ['caerus', 'mask', 'topology', None]
        filters = ['lee', 'lee_enhanced', 'lee_sigma', 'kuan', 'fastnl', 'bilateral', 'frost', 'median', 'mean', None]
        windows = [None, 3, 63]
        cus = [None, 0, 0.75]
        img = fp.import_example('2dpeaks')

        for getfilter in filters:
            for window in windows:
                for cu in cus:
                    fp = findpeaks(method='topology', imsize=None, scale=True, togray=True, denoise=getfilter, params={'window': window, 'cu': cu})
                    assert fp.fit(img)
                    # assert fp.plot_mesh(wireframe=False)
                    # plt.close('all')
                    # assert fp.plot_persistence()
                    # plt.close('all')
                    # assert fp.plot()
                    # plt.close('all')
                    # assert fp.plot_preprocessing()
  
    def test_persistence_plot_parameters(self):
        """Test persistence plotting parameters"""
        fp = findpeaks(method='topology')
        X = fp.import_example('1dpeaks')
        results = fp.fit(X)
        
        # Test different figsize options
        figsize_options = [(10, 5), (20, 10), (15, 8)]
        for figsize in figsize_options:
            fp.plot_persistence(figsize=figsize)
        
        # Test different fontsize options
        fontsize_options = [10, 14, 18, 20]
        for fontsize in fontsize_options:
            fp.plot_persistence(fontsize_ax1=fontsize, fontsize_ax2=fontsize)
        
        # Test different label options
        label_options = ['x-axis', 'y-axis', 'Time', 'Value']
        for label in label_options:
            fp.plot_persistence(xlabel=label, ylabel=label)
    
    def test_plot_preprocessing(self):
        """Test preprocessing plot functionality"""
        fp = findpeaks(method='topology', denoise='lee')
        X = fp.import_example('2dpeaks')
        results = fp.fit(X)
        
        # Test that preprocessing plot works
        fp.plot_preprocessing()
    
    def test_plot_mask(self):
        """Test mask plotting functionality"""
        fp = findpeaks(method='mask')
        X = fp.import_example('2dpeaks')
        results = fp.fit(X)
        
        # Test different mask plot parameters
        limit_options = [0, 0.1, 0.5, None]
        for limit in limit_options:
            ax = fp.plot_mask(limit=limit)
        
        # Test different figure_order options
        for order in ['vertical', 'horizontal']:
            ax = fp.plot_mask(figure_order=order)
        
        # Test different marker and color options
        fp.plot_mask(marker='o', color='blue', s=5, fontsize=10)
    
    def test_caerus_method(self):
        """Test caerus method specifically"""
        fp = findpeaks(method='caerus')
        X = fp.import_example('1dpeaks')
        
        # Test with default parameters
        results = fp.fit(X)
        assert results is not None
        
        # Test with custom parameters
        fp = findpeaks(method='caerus', params={'window': 5, 'delta': 0.1})
        results = fp.fit(X)
        assert results is not None
    
    def test_2d_with_different_methods(self):
        """Test 2D data with different methods"""
        fp = findpeaks()
        X = fp.import_example('2dpeaks')
        
        # Test topology method
        fp = findpeaks(method='topology')
        results = fp.fit(X)
        assert results is not None
        assert 'persistence' in results
        
        # Test mask method
        fp = findpeaks(method='mask')
        results = fp.fit(X)
        assert results is not None
        assert 'Xdetect' in results
        
    def test_1d_with_different_methods(self):
        """Test 1D data with different methods"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        
        # Test topology method
        fp = findpeaks(method='topology')
        results = fp.fit(X)
        assert results is not None
        assert 'persistence' in results
        
        # Test peakdetect method
        fp = findpeaks(method='peakdetect')
        results = fp.fit(X)
        assert results is not None
        assert 'df' in results
        
        # Test caerus method
        fp = findpeaks(method='caerus')
        results = fp.fit(X)
        assert results is not None
    
    def test_noise_handling(self):
        """Test handling of noisy data"""
        import numpy as np
        
        # Create noisy data
        np.random.seed(42)
        x = np.linspace(0, 10, 100)
        y = np.sin(x) + 0.1 * np.random.randn(100)
        
        # Test with different denoising methods
        denoise_methods = ['fastnl', 'bilateral', 'lee', 'median', 'mean']
        
        for method in denoise_methods:
            fp = findpeaks(method='topology', denoise=method)
            results = fp.fit(y)
            assert results is not None
    
    def test_large_dataset_handling(self):
        """Test handling of larger datasets"""
        import numpy as np
        
        # Create larger dataset
        np.random.seed(42)
        X_large = np.random.randn(1000)
        
        # Test with different methods
        methods = ['topology', 'peakdetect', 'caerus']
        
        for method in methods:
            fp = findpeaks(method=method)
            results = fp.fit(X_large)
            assert results is not None
    
    def test_parameter_validation(self):
        """Test parameter validation and error handling"""
        X = [10, 11, 9, 23, 21, 11, 45, 20, 11, 12]
        
        # Test invalid method
        try:
            fp = findpeaks(method='invalid_method')
            results = fp.fit(X)
            # Should either work with default method or raise an exception
        except:
            pass  # Expected for invalid method
        
        # Test invalid parameters
        try:
            fp = findpeaks(method='peakdetect', lookahead=-1)
            results = fp.fit(X)
            # Should either work or raise an exception
        except:
            pass  # Expected for negative lookahead
        
        # Test invalid interpolation factor
        try:
            fp = findpeaks(method='peakdetect', interpolate=0)
            results = fp.fit(X)
            # Should either work or raise an exception
        except:
            pass  # Expected for zero interpolation
    
    def test_specific_error_case(self):
        """Test the specific case that was causing the OpenCV error"""
        fp = findpeaks()
        X = fp.import_example('2dpeaks')
        
        # This was the problematic case: scale=False with fastnl denoising
        fp = findpeaks(method='topology', scale=False, denoise='fastnl')
        results = fp.fit(X)
        assert results is not None
        
        # Also test with bilateral denoising
        fp = findpeaks(method='topology', scale=False, denoise='bilateral')
        results = fp.fit(X)
        assert results is not None


# ----------------------------------------------------
# Name        : interpolate.py
# Author      : E.Taskesen
# Contact     : erdogant@gmail.com
# Licence     : MIT
# ----------------------------------------------------

import numpy as np
from scipy.interpolate import make_interp_spline, interp1d
import matplotlib.pyplot as plt
import logging

logger = logging.getLogger(__name__)

# %%
def interpolate_line1d(X, n=3, method=2, showfig=False):
    """Interpolate 1d-vector.

    Parameters
    ----------
    X : array-like (1D-vector)
        Input image data.
    n : int, (default : 3)
        The interpolation factor. The data is interpolation by a factor n.
    method : str (default: 'linear')
         String or integer
         * 0 : order degree
         * 1 : order degree
         * 2 : order degree
         * 3 : order degree
         * 'linear' (default)
         * 'nearest'
         * 'zero'
         * 'slinear'
         * 'quadratic'
         * 'cubic'
         * 'previous'
         * 'next'
        showfig : bool, (default : False)
            Show the figure.

    Returns
    -------
    X : array-like (1D-vector)
        Interpolated data.

    """

    nboost = len(X) * n
    if len(X)>nboost: raise Exception('[findpeaks] >nboost (n=%.0f) must be larger then input data (n=%.0f)' %(nboost, len(X)))
    bootstdata = np.zeros(nboost) * np.nan
    idx = np.unique(np.floor(np.linspace(0, len(bootstdata) - 1, len(X))).astype(int))
    bootstdata[idx] = X

    X = interpolate_nans(bootstdata, method=method)
    logger.info('Interpolating 1d-vector by factor %d' %(n))

    if showfig:
        plot(X, bootstdata, method)

    return(X)

# %% Smooting of the line
def interpolate_line2d(xs, ys=None, interpol=3, window=3):
    """interpolate 2D vector.

    2D Interpolation Description
    ----------------------------
    Smoothing a 2d vector can be challanging if the data is low sampled.
    This function contains two steps. First interpolation of the input line followed by a convolution.

    Parameters
    ----------
    xs : array-like
        Data points for the x-axis.
    ys : array-like
        Data points for the y-axis.
    interpol : int, (default : 3)
        The interpolation factor. The data is interpolation by a factor n before the smoothing step.
    window : int, (default : 3)
        Smoothing window that is used to create the convolution and gradually smoothen the line.

    Returns
    -------
    xnew : array-like
        Data points for the x-axis.
    ynew : array-like
        Data points for the y-axis.

    """
    if window is not None:
        logger.info('Interpolating 2d-array (image) by factor %d' %(interpol))
        # Specify number of points to interpolate the data
        # Interpolate xs line
        extpoints = np.linspace(0, len(xs), len(xs) * interpol)
        spl = make_interp_spline(range(0, len(xs)), xs, k=3)
        xnew = spl(extpoints)
        xnew[window:-window]

        # First smoothing on the raw input data
        ynew=None
        if ys is not None:
            ys = _smooth(ys, window)
            # Interpolate ys line
            spl = make_interp_spline(range(0, len(ys)), ys, k=3)
            ynew = spl(extpoints)
            ynew[window:-window]
    else:
        xnew, ynew = xs, ys
    return xnew, ynew

def _smooth(X, window):
    box = np.ones(window) / window
    X_smooth = np.convolve(X, box, mode='same')
    return X_smooth

# %% Plot
def plot(X, bootstdata, method):
    plt.figure()
    # plt.plot(X, label='Boosted')
    plt.plot(interpolate_nans(bootstdata, method=0), label='0nd order')
    plt.plot(interpolate_nans(bootstdata, method=1), label='1nd order')
    plt.plot(interpolate_nans(bootstdata, method=2), label='2nd order')
    plt.plot(interpolate_nans(bootstdata, method=3), label='3nd order')
    plt.grid(True)
    plt.legend()


# %% interpolatie
def interpolate_nans(X, method='linear', replace_value_to_nan=None):
    """Interpolate the nan values in an 1D array.

    Parameters
    ----------
    X : array-like
        input data X.
    replace_value_to_nan : float (default: None)
        Replace value to np.nan.
        * None : take finite data: interplate np.nan.
        * 0 : take np.nan with the additional value (alternative)
    method : str (default: 'linear')
         String or integer
         * 0 : order degree
         * 1 : order degree
         * 2 : order degree
         * 3 : order degree
         * 'linear' (default)
         * 'nearest'
         * 'zero'
         * 'slinear'
         * 'quadratic'
         * 'cubic'
         * 'previous'
         * 'next'

    Returns
    -------
    None.

    Examples
    --------
    >>> X = np.array([1,2,3,np.nan,np.nan,6,7,np.nan,9,10])
    >>> Xint1 = interpolate_nans(X)
    >>> Xint2 = interpolate_nans(X, method=0)

    """
    yhat = []
    # Replace replace_value_to_nan by NaN value
    if replace_value_to_nan is not None: 
        X[X==replace_value_to_nan]=np.nan
    good = np.where(np.isfinite(X))[0]

    # interpolate to fill nan values
    inds = np.arange(X.shape[0])

    # Check for nan values
    if len(good) == 0:
        logger.warning('WARNING: Skipping because nothing to do: No nan values (?)')
        # yhat= np.nan_to_num(X)
        yhat = X
        # np.nan_to_num(X)
    else:
        f = interp1d(inds[good], X[good], bounds_error=False, kind=method)
        yhat = np.where(np.isfinite(X), X, f(inds))

    return(yhat)


# ----------------------------------------------------
# Name: findpeaks.py
# Author: E.Taskesen
# Contact: erdogant@gmail.com
# github: https://github.com/erdogant/findpeaks
# Licence: MIT
# ----------------------------------------------------

try:
    import findpeaks.union_find as union_find
    from findpeaks.filters.lee import lee_filter
    from findpeaks.filters.lee_enhanced import lee_enhanced_filter
    from findpeaks.filters.lee_sigma import lee_sigma_filter
    from findpeaks.filters.kuan import kuan_filter
    from findpeaks.filters.frost import frost_filter
    from findpeaks.filters.median import median_filter
    from findpeaks.filters.mean import mean_filter
except:
    #### DEBUG ONLY ####
    import union_find as union_find
    from filters.lee import lee_filter
    from filters.lee_enhanced import lee_enhanced_filter
    from filters.lee_sigma import lee_sigma_filter
    from filters.kuan import kuan_filter
    from filters.frost import frost_filter
    from filters.median import median_filter
    from filters.mean import mean_filter
# ######################

from scipy.ndimage.morphology import generate_binary_structure, binary_erosion
from scipy.ndimage.filters import maximum_filter

from tqdm import tqdm
import numpy as np
import pandas as pd
import logging

logger = logging.getLogger(__name__)


# %% Import cv2
def _import_cv2():
    # Only for 2D images required
    try:
        import cv2
        return cv2
    except:
        raise ImportError('cv2 must be installed manually. Try to: <pip install opencv-python> or <pip install opencv-python-headless>')


# %% Scaling
def scale(X):
    """Normalize data (image) by scaling.

    Scaling Description
    -------------------
    Scaling in range [0-255] by img*(255/max(img))

    Parameters
    ----------
    X: array-like
        Input image data.

    Returns
    -------
    X: array-like
        Scaled image.

    """
    logger.info('Scaling image between [0-255] and to uint8')
    try:
        # Normalizing between 0-255
        X = X - X.min()
        try:
            X = X / X.max()
        except:
            X = X * 0
        X = X * 255
        # Downscale typing
        X = np.uint8(X)
    except:
        logger.warning('Scaling not possible.')
    return X


# %%
def togray(X):
    """Convert color to grey-image.

    Grayscale Conversion Description
    --------------------------------
    Convert 3d-RGB colors to 2d-grey image.

    Parameters
    ----------
    X: array-like
        Input image data.

    Returns
    -------
    X: array-like
        2d-image.

    """
    # Import cv2
    cv2 = _import_cv2()
    try:
        logger.info('Conversion to gray image.')
        
        # Check if image is already grayscale (1 channel)
        if len(X.shape) == 2:
            logger.info('Image is already grayscale, no conversion needed.')
            return X
        
        # Check if image has valid number of channels for conversion
        if len(X.shape) != 3 or X.shape[2] not in [3, 4]:
            logger.warning(f'Image has {len(X.shape)} dimensions with shape {X.shape}, cannot convert to grayscale.')
            return X
            
        if X.dtype != np.uint8:
            logger.warning(f'Input image dtype is {X.dtype}, converting to uint8.')
            X = X.astype(np.uint8)

        X = cv2.cvtColor(X, cv2.COLOR_BGR2GRAY)
    except Exception as e:
        logger.error(f'Conversion to gray failed using cv2: {e}')
        # Return original image if conversion fails
        return X

    return X


# %%
def resize(X, size=None):
    """Resize image.

    Parameters
    ----------
    X: array-like
        Input image data.
    size: tuple, (default: None)
        size to desired (width,length).

    Returns
    -------
    X: array-like

    """
    # Import cv2
    cv2 = _import_cv2()
    try:
        if size is not None:
            logger.info('Resizing image to %s.' %(str(size)))
            X = cv2.resize(X, size)
    except Exception as e:
        logger.warning(f'Resizing not possible: {e}')
    return X


# %%
def denoise(X, method='fastnl', window=9, cu=0.25):
    """Denoise input data.

    Denoising Description
    ---------------------
    Denoising the data is very usefull before detection of peaks. Multiple methods are implemented to denoise the data.
    The bilateral filter uses a Gaussian filter in the space domain,
    but it also uses one more (multiplicative) Gaussian filter component which is a function of pixel intensity differences.
    The Gaussian function of space makes sure that only pixels are 'spatial neighbors' are considered for filtering,
    while the Gaussian component applied in the intensity domain (a Gaussian function of intensity differences)
    ensures that only those pixels with intensities similar to that of the central pixel ('intensity neighbors')
    are included to compute the blurred intensity value. As a result, this method preserves edges, since for pixels lying near edges,
    neighboring pixels placed on the other side of the edge, and therefore exhibiting large intensity variations when
    compared to the central pixel, will not be included for blurring.

    Parameters
    ----------
    X: array-like
        Input image data.
    method: string, (default: 'fastnl', None to disable)
        Filtering method to remove noise
            * None
            * 'fastnl'
            * 'bilateral'
            * 'lee'
            * 'lee_enhanced'
            * 'lee_sigma'
            * 'kuan'
            * 'frost'
            * 'median'
            * 'mean'
    window: int, (default: 3)
        Denoising window. Increasing the window size may removes noise better but may also removes details of image in certain denoising methods.
    cu: float, (default: 0.25)
        The noise variation coefficient, applies for methods: ['kuan','lee','lee_enhanced']
    sigma: float, (default: 0.9)
        Speckle noise standard deviation, applies for methods: ['lee_sigma']
    num_looks : int, (default: 1)
        Number of looks of the SAR img, applies for methods: ['lee_sigma']
    tk: int, (default: 5)
        Threshold of neighbouring pixels outside of the 98th percentile, applies for methods: ['lee_sigma']

    Returns
    -------
    X: array-like
        Denoised data.

    References
    ----------
    * https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_imgproc/py_filtering/py_filtering.html

    """
    if window is None: window=9
    if cu is None: cu=0.25
    # Import library
    cv2 = _import_cv2()

    # Peform the denoising
    # try:
    logger.info('Denoising with [%s], window: [%d].' %(method, window))
    if method=='fastnl':
        if len(X.shape)==2:
            X = cv2.fastNlMeansDenoising(X, h=window)
        elif len(X.shape)==3:
            logger.info('Denoising color image.')
            X = cv2.fastNlMeansDenoisingColored(X, h=window)
        else:
            logger.warning(f'Cannot denoise image with {len(X.shape)} dimensions')
    elif method=='bilateral':
        X = cv2.bilateralFilter(X, window, 75, 75)
    elif method=='lee':
        X = lee_filter(X, win_size=window, cu=cu)
    elif method=='lee_enhanced':
        X = lee_enhanced_filter(X, win_size=window, cu=cu, k=1, cmax=1.73)
    elif method=='lee_sigma':
        X = lee_sigma_filter(X, sigma=0.9, win_size=window, num_looks=1, tk=5)
    elif method=='kuan':
        X = kuan_filter(X, win_size=window, cu=cu)
    elif method=='frost':
        X = frost_filter(X, win_size=window, damping_factor=2)
    elif method=='median':
        X = median_filter(X, win_size=window)
    elif method=='mean':
        X = mean_filter(X, win_size=window)
    # except:
    #     if verbose>=2: print('[findpeaks] >Warning: Denoising failt!')
    return X


# %%
def mask(X, limit=0):
    """Determine peaks in 2d-array using a mask.

    Mask Method Description
    -----------------------
    Takes an image and detect the peaks using the local maximum filter.
    Returns a boolean mask of the peaks (i.e. 1 when the pixel's value is the neighborhood maximum, 0 otherwise)

    Parameters
    ----------
    X: array-like
        Input image data.
    limit: float, (default: None)
        Values > limit are set as regions of interest (ROI).

    Returns
    -------
    dict()
        Xraw: array-like.
            Input image.
        Xdetect: array-like (same shape as input data)
            detected peaks with respect the input image. Elements are the scores.
        Xranked: array-like (same shape as input data)
            detected peaks with respect the input image. Elements are the ranked peaks (1=best).

    References
    ----------
    * https://stackoverflow.com/questions/3684484/peak-detection-in-a-2d-array

    """
    if limit is None: limit=0

    logger.info('Detect peaks using the mask method with limit=%s.' %(limit))
    # define an 8-connected neighborhood
    neighborhood = generate_binary_structure(2, 2)

    # apply the local maximum filter; all pixel of maximal value in their neighborhood are set to 1
    local_max = maximum_filter(X, footprint=neighborhood)==X
    # local_max is a mask that contains the peaks we are looking for, but also the background.
    # In order to isolate the peaks we must remove the background from the mask.

    # we create the mask of the background
    background = (X <= limit)

    # Erode the background in order to successfully subtract it form local_max,
    # otherwise a line will appear along the background border (artifact of the local maximum filter)
    eroded_background = binary_erosion(background, structure=neighborhood, border_value=1)

    # We obtain the final mask, containing only peaks, by removing the background from the local_max mask (xor operation)
    Xdetect = local_max ^ eroded_background

    # Nothing to rank but for consistency purposes
    Xranked = Xdetect.astype(int)
    idxs=np.where(Xranked)
    for i, idx in enumerate(zip(idxs[0], idxs[1])):
        Xranked[idx]=i+1

    # Make dict
    result = {'Xdetect': Xdetect, 'Xranked': Xranked}

    # Return
    return result


# %%
def topology2d(X, limit=None, whitelist=['peak','valley']):
    """Determine peaks and valleys in 2d-array using toplogy method.

    Topology 2D Description
    -----------------------
    This function calls the topology function that ONLY computes peaks in case of 2d-arrays.
    To detect the valleys, the image is inverted and the topology function is called.
    The final results is the combined peak and valleys.

    Parameters
    ----------
    X: array-like data
        Input data.
    limit: float, (default: None)
        score > limit are set as regions of interest (ROI).

    Returns
    -------
    dict()
        Xdetect: array-like
            detected peaks/valleys in the same shape as the input image. Elements are the scores. (high=best peak and low=best valley).
        Xranked: array-like
            detected peaks/valleys in the same shape as the input image. Elements are the ranked peaks (1=best peak and -1 is best valley).
        persistence: DataFrame().
            * x, y: Coordinates
            * birth: Birth level, tuple(coordinate, rgb value)
            * death: Death level, tuple(coordinate, rgb value)
            * score: Persistence scores
            * peak: True if peak
            * valley: True if valley

    """
    result_peak = {'groups0': [], 'Xdetect': np.zeros_like(X).astype(float), 'Xranked': np.zeros_like(X).astype(float), 'peak': None, 'valley': None, 'persistence': pd.DataFrame(columns=['x','y','birth_level','death_level','score'])}
    result_valley = {'groups0': [], 'Xdetect': np.zeros_like(X).astype(float), 'Xranked': np.zeros_like(X).astype(float), 'peak': None, 'valley': None, 'persistence': pd.DataFrame(columns=['x','y','birth_level','death_level','score'])}

    # Detect peaks
    if np.any(np.isin(whitelist, 'peak')):
        result_peak = topology(X, limit=limit, reverse=True)
        result_peak['persistence']['peak']=True
        result_peak['persistence']['valley']=False

    # Compute max value in array to create the negative image
    # max_val = 255 if np.max(X.ravel())>1 else 1
    # Detect valleys
    # result_valley = topology(max_val-X, limit=limit)
    if np.any(np.isin(whitelist, 'valley')):
        result_valley = topology(X, limit=limit, reverse=False)
        result_valley['persistence']['peak']=False
        result_valley['persistence']['valley']=True
        result_valley['persistence']['death_level'] = result_valley['persistence']['death_level'] * -1
        result_valley['persistence']['birth_level'] = result_valley['persistence']['birth_level'] * -1

    # Combine results
    persistence = pd.concat([result_peak['persistence'], result_valley['persistence']])
    persistence.sort_values(by='score', ascending=False, inplace=True)
    # Combine datamatrix
    Xdetect = result_peak['Xdetect']
    idx = np.where(result_valley['Xdetect'])
    if len(idx[0])>0: Xdetect[idx] = result_valley['Xdetect'][idx] * -1
    # Combine the ranked ones
    Xranked = result_peak['Xranked']
    idx = np.where(result_valley['Xranked'])
    if len(idx[0])>0: Xranked[idx] = result_valley['Xranked'][idx] * -1
    # Combine the group0
    groups0 = result_peak['groups0'] + result_valley['groups0']
    # Make dict
    result = {'persistence': persistence, 'Xdetect': Xdetect, 'Xranked': Xranked, 'groups0': groups0}
    # Return
    return result


# %%
def topology(X, limit=None, reverse=True, neighborhood_generator=None):
    """Determine peaks using topology method based on persistent homology.

    Topology Method Description
    ---------------------------
    The topology method uses persistent homology to detect peaks and valleys in data.
    The core idea is to consider the function graph that assigns each pixel its level,
    then simulate a water level that continuously descends to lower levels.

    At local maxima, islands emerge (birth events). At saddle points, two islands merge;
    the lower island merges into the higher island (death events). The persistence diagram
    depicts death-over-birth values of all islands, where persistence is the difference
    between birth and death levels - the vertical distance from the main diagonal.

    **Advantages over thresholding methods:**

    - Naturally handles noise and artifacts that affect simple thresholding
    - Provides quantitative significance measures (persistence scores)
    - Mathematically stable and robust
    - Enables principled filtering based on persistence rather than arbitrary thresholds

    **Applications:**

    - **Line Detection**: Particularly effective for Hough Transform applications where
      traditional voting methods are susceptible to noise. The persistence-based approach
      significantly outperforms threshold-based methods in synthetic data experiments.
    - **Peak Detection**: General peak/valley detection in 1D and 2D data
    - **Feature Detection**: Robust detection of significant features in noisy data

    **Hough Transform Integration:**

    This method addresses key limitations of classical Hough transform voting:

    - Replaces simple thresholding with persistent homology-based peak detection
    - Enhances robustness against noise and artifacts
    - Provides mathematical stability guarantees
    - Enables principled parameter selection based on persistence scores

    The method not only identifies local maxima but also quantifies their significance
    through persistence scores, allowing for principled filtering of peaks based on
    their topological importance rather than arbitrary thresholds.

    Parameters
    ----------
    X: array-like data
        Input data.
    limit: float, (default: None)
        score > limit are set as regions of interest (ROI).
        None: Take all positions into consideration.
    reverse: bool, (default: True)
        For 1d-vectors, reverse should be True to detect peaks and valleys.
        For 2d-images, the peaks are detected by setting reverse=True and valleys by reverse=False.
    neighborhood_generator: callable, (default: None)
        Custom neighborhood generator function. If None, uses default 8-neighborhood.
        The function should take parameters (p, w, h) and return neighboring coordinates.

    Returns
    -------
    dict()
        Xdetect: array-like
            detected peaks in the same shape as the input image. Elements are the scores (higher is better).
        Xranked: array-like
            detected peaks in the same shape as the input image. Elements are the ranked peaks (1=best).
        peak: array-like
            Detected peaks
        valley: array-like
            Detected valley
        groups0: array-like
            Unstructured results
        persistence: DataFrame()
            DataFrame with columns:

            - x, y: Coordinates
            - birth: Birth level, tuple(coordinate, rgb value)
            - death: Death level, tuple(coordinate, rgb value)
            - score: Persistence scores

    References
    ----------
        * Johannes Ferner et al, Persistence-based Hough Transform for Line Detection, https://arxiv.org/abs/2504.16114
        * H. Edelsbrunner and J. Harer, Computational Topology. An Introduction, 2010, ISBN 0-8218-4925-5.
        * Initial implementation: Stefan Huber <shuber@sthu.org>, editted by: Erdogan Taskesen <erdogant@gmail.com>, 2020

    """
    if limit is None:
        limit = float(np.min(X)) - 1
    if X.max().max()<limit:
        limit=X.max().max()-1
        logger.info('Minimum limit should be %s or smaller.' %(limit))

    logger.info('Detect peaks using topology method with limit at %s.' %(limit))

    if not reverse:
        X = reverse_values(X.copy())
        reverse=True

    results = {'groups0': [], 'Xdetect': np.zeros_like(X).astype(float), 'Xranked': np.zeros_like(X).astype(float), 'peak': None, 'valley': None, 'persistence': pd.DataFrame(columns=['x','y','birth_level','death_level','score'])}
    h, w = X.shape
    max_peaks, min_peaks = None, None
    groups0 = {}

    # It is important to ensure unique values because the method sorts the values and only unique values are processed.
    # Without adjusting duplicated values, peaks with exactly the same height will be skipped.
    if X.shape[1] <=2:
        X = _make_unique(X)
    else:
        X = _make_unique_fast(X)
        # X = np.maximum(X + ((X > 0).astype(int) * np.random.random(X.shape) / 10), 0)

    # Get indices orderd by value from high to low. As a tie-breaker, we use
    indices = list(zip(*np.where(X >= limit)))

    # We add p as a secondary key to have an unambiguous total order below when
    # we enumerate neighboring cells of cells and consistency regarding
    # "oldest" component.
    indices.sort(key=lambda p: (_get_indices(X, p), p), reverse=reverse)

    # Maintains the growing sets
    uf = union_find.UnionFind()

    def _get_comp_birth(p):
        return _get_indices(X, uf[p])

    # Process pixels from high to low
    for i, p in tqdm(enumerate(indices), disable=disable_tqdm(), desc=logger.info("Topology")):
        v = _get_indices(X, p)

        if neighborhood_generator is None:
            ni = [uf[q] for q in _iter_neighbors(p, w, h) if q in uf]
        else:
            ni = [uf[q] for q in _iter_neighbors_generator(p, w, h)(p, w, h, neighborhood_generator=neighborhood_generator) if q in uf]

        # Sort by (value, index) as key. Note that this is the same sorting order as for indices. Otherwise, we have an inconsistent notion of the "older" component!
        nc = sorted([(_get_comp_birth(q), q) for q in set(ni)], reverse=True)

        if i == 0:
            groups0[p] = (v, v, None)

        uf.add(p, -i)

        if len(nc) > 0:
            oldp = nc[0][1]
            uf.union(oldp, p)
            # Merge all others with oldp
            for bl, q in nc[1:]:
                if uf[q] not in groups0:
                    groups0[uf[q]] = (float(bl), float(bl) - float(v), p)
                uf.union(oldp, q)

    groups0 = [(k, groups0[k][0], groups0[k][1], groups0[k][2]) for k in groups0]
    groups0.sort(key=lambda g: g[2], reverse=True)

    # Filter on limit
    if (limit is not None):
        Ikeep = np.array(list(map(lambda x: x[2], groups0))) > limit
        groups0 = np.array(groups0, dtype='object')
        groups0 = groups0[Ikeep].tolist()

    if len(groups0)>0:
        # Extract the max peaks and sort
        max_peaks = np.array(list(map(lambda x: [x[0][0], x[1]], groups0)))
        idxsort = np.argsort(max_peaks[:, 0])
        max_peaks = max_peaks[idxsort, :]
        # Extract the min peaks and sort
        min_peaks = np.array(list(map(lambda x: [(x[3][0] if x[3] is not None else 0), x[2]], groups0)))
        idxsort = np.argsort(min_peaks[:, 0])
        min_peaks = min_peaks[idxsort, :]
        # min_peaks = np.round(min_peaks, decimals=3)

        # Build the output results in the same manner as the input image
        Xdetect = np.zeros_like(X).astype(float)
        Xranked = np.zeros_like(X).astype(int)
        for i, homclass in enumerate(groups0):
            p_birth, bl, pers, p_death = homclass
            y, x = p_birth
            Xdetect[y, x] = pers
            Xranked[y, x] = i + 1

        # If data is 1d-vector (and not image), make single vector
        if (X.shape[1]==2): #and (np.all(Xdetect[:, 1]==0)):
            Xdetect = Xdetect[:, 0]
            Xranked = Xranked[:, 0]

        # Store in dataframe
        df_persistence = pd.DataFrame()
        df_persistence['x'] = np.array(list(map(lambda x: x[0][1], groups0)))
        df_persistence['y'] = np.array(list(map(lambda x: x[0][0], groups0)))
        df_persistence['birth_level'] = np.array(list(map(lambda x: float(x[1]), groups0)))
        df_persistence['death_level'] = np.array(list(map(lambda x: float(x[1]) - float(x[2]), groups0)))
        df_persistence['score'] = np.array(list(map(lambda x: float(x[2]), groups0)))
        # Results
        results = {}
        results['groups0'] = groups0
        results['Xdetect'] = Xdetect
        results['Xranked'] = Xranked
        results['peak'] = max_peaks
        results['valley'] = min_peaks
        results['persistence'] = df_persistence

    # return
    return results


def reverse_values(image_array):
    # Clip the values to ensure they are between 0 and 255
    clipped_array = np.clip(image_array, np.min(image_array), np.max(image_array))

    # Reverse the values
    reversed_array = np.max(image_array) - clipped_array

    return reversed_array

def _get_indices(im, p):
    # return im[p[0]][p[1]]
    return im[p[0], p[1]]

def generate_default_neighborhood(p, h, w, eight_neighborship=True):
    y, x = p

    if eight_neighborship:
        neigh = [(y + j, x + i) for i in [-1, 0, 1] for j in [-1, 0, 1]]
    else:
        # 4-neighborship
         neigh = [(y-1, x), (y+1, x), (y, x-1), (y, x+1)]

    neigh = [(j, i) for j, i in neigh if 0 <= j < h or 0 <= i < w or (j != y or i != x)]

    return neigh


def _iter_neighbors_generator(p, w, h, neighborhood_generator=None):
    if neighborhood_generator is None:
        neighborhood_generator = generate_default_neighborhood

    neigh = neighborhood_generator(p, w, h)

    for j, i in neigh:
        yield j, i


def _iter_neighbors(p, w, h):
    y, x = p

    # 8-neighborship
    neigh = [(y + j, x + i) for i in [-1, 0, 1] for j in [-1, 0, 1]]
    # 4-neighborship
    # neigh = [(y-1, x), (y+1, x), (y, x-1), (y, x+1)]

    for j, i in neigh:
        if j < 0 or j >= h:
            continue
        if i < 0 or i >= w:
            continue
        if j == y and i == x:
            continue
        yield j, i

def _post_processing(X, Xraw, min_peaks, max_peaks, interpolate, lookahead, labxRaw=None):
    if lookahead<1: raise Exception('[findpeaks] >lookhead parameter should be at least 1.')
    labx_s = np.zeros((len(X))) * np.nan
    results = {}
    results['min_peaks_s'] = None
    results['max_peaks_s'] = None
    results['xs'] = np.arange(0, len(Xraw))
    results['labx_s'] = labx_s
    results['labx'] = np.zeros((len(Xraw))) * np.nan
    results['min_peaks'] = None
    results['max_peaks'] = None

    if len(min_peaks)>0 and len(max_peaks)>0 and (max_peaks[0][0] is not None):

        idx_peaks, _ = zip(*max_peaks)
        idx_peaks = np.array(list(idx_peaks)).astype(int)
        idx_valleys, _ = zip(*min_peaks)
        # Add first and last row
        # idx_valleys = np.array(idx_valleys)
        idx_valleys = np.append(np.array(list(idx_valleys)), len(X) - 1).astype(int)
        idx_valleys = np.append(0, idx_valleys)

        # Group distribution
        count=1
        for i in range(0, len(idx_valleys) - 1):
            if idx_valleys[i]!=idx_valleys[i + 1]:
                labx_s[idx_valleys[i]:idx_valleys[i + 1] + 1] = count
                count=count + 1

        # Scale back to original data
        if interpolate is not None:
            min_peaks = np.minimum(np.ceil(((idx_valleys / len(X)) * len(Xraw))).astype(int), len(Xraw) - 1)
            max_peaks = np.minimum(np.ceil(((idx_peaks / len(X)) * len(Xraw))).astype(int), len(Xraw) - 1)
            # Scaling is not accurate for indexing and therefore, a second wave of searching for max_peaks
            max_peaks_corr = []
            for max_peak in max_peaks:
                getrange = np.arange(np.maximum(max_peak - lookahead, 0), np.minimum(max_peak + lookahead, len(Xraw)))
                max_peaks_corr.append(getrange[np.argmax(Xraw[getrange])])
            # Scaling is not accurate for indexing and therefore, a second wave of searching for min_peaks
            min_peaks_corr = []
            for min_peak in min_peaks:
                getrange = np.arange(np.maximum(min_peak - lookahead, 0), np.minimum(min_peak + lookahead, len(Xraw)))
                min_peaks_corr.append(getrange[np.argmin(Xraw[getrange])])
            # Set the labels
            count = 1
            labx = np.zeros((len(Xraw))) * np.nan
            for i in range(0, len(min_peaks) - 1):
                if min_peaks[i]!=min_peaks[i + 1]:
                    labx[min_peaks[i]:min_peaks[i + 1] + 1] = count
                    count=count + 1

            # Store based on original
            results['labx'] = labx
            results['min_peaks'] = np.c_[min_peaks_corr, Xraw[min_peaks_corr]]
            results['max_peaks'] = np.c_[max_peaks_corr, Xraw[max_peaks_corr]]

        results['min_peaks_s'] = np.c_[idx_valleys, X[idx_valleys]]
        results['max_peaks_s'] = np.c_[idx_peaks, X[idx_peaks]]
        if labxRaw is None:
            results['labx_s'] = labx_s
        else:
            results['labx_s'] = labxRaw

    # Return
    return results

# %% Normalize.
def normalize(X, minscale = 0.5, maxscale = 4, scaler: str = 'zscore'):
    # Instead of Min-Max scaling, that shrinks any distribution in the [0, 1] interval, scaling the variables to
    # Z-scores is better. Min-Max Scaling is too sensitive to outlier observations and generates unseen problems,

    # Set sizes to 0 if not available
    X[np.isinf(X)]=0
    X[np.isnan(X)]=0
    if minscale is None: minscale=0.5

    # out-of-scale datapoints.
    if scaler == 'zscore' and len(np.unique(X)) > 3:
        # Handle division by zero or very small values to prevent RuntimeWarning
        X_std = np.std(X)
        if abs(X_std) < 1e-10:  # Very small threshold to avoid numerical issues
            X = X.flatten() - np.mean(X)  # Skip division when std is essentially zero
        else:
            X = (X.flatten() - np.mean(X)) / X_std
        X = X + (minscale - np.min(X))
    elif scaler == 'minmax':
        min_val = np.min(X)
        max_val = np.max(X)
        X = minscale + maxscale * (X - min_val) / (max_val - min_val)
    else:
        X = X.ravel()
    # Max digits is 4
    X = np.array(list(map(lambda x: round(x, 4), X)))

    return X

# %%
def disable_tqdm():
    """Set the logger for verbosity messages."""
    return (True if (logger.getEffectiveLevel()>=30) else False)

def _make_unique_fast(X: np.ndarray) -> np.ndarray:
    """Add small deterministic noise to make values unique, preserving relative order."""
    if not np.issubdtype(X.dtype, np.floating):
        X = X.astype(np.float64)
    eps = np.finfo(X.dtype).eps
    noise = np.arange(X.size, dtype=np.float64).reshape(X.shape) * eps

    # Only apply noise where X is not exactly zero
    return np.where(X != 0, X + noise, X)

def _make_unique(arr: np.ndarray):
    """Method iterates through elements of the input array to ensure all values are unique.
       Duplicate values are reduced by the smallest possible increment for the given data type.
       """
    logger.debug('Making values unique')
    res = np.empty_like(arr)
    it = np.nditer([arr, res], [], [['readonly'], ['writeonly', 'allocate']])
    seen = set()
    with it:
        while not it.finished:
            a = it[0].item()
            while a in seen and np.isfinite(a):
                a = np.nextafter(a, -np.inf)
            it[1] = a
            if a not in seen:
                seen.add(a)
            it.iternext()
    return res

# def _make_unique(arr: np.ndarray) -> np.ndarray:
#     """Make all finite values in the array unique by perturbing duplicates slightly."""
#     # Convert to float64 if needed
#     if not np.issubdtype(arr.dtype, np.floating):
#         arr = arr.astype(np.float64)
#     dtype = arr.dtype

#     arr_flat = arr.ravel()
#     res = arr_flat.copy()
#     is_finite = np.isfinite(res)
#     finite_vals = res[is_finite]
#     unique, counts = np.unique(finite_vals, return_counts=True)
#     dupes = unique[counts > 1]
#     if dupes.size == 0:
#         return arr.copy()
    
#     seen = {}
#     mask = np.isin(finite_vals, dupes)
#     indices = np.where(mask)[0]
#     values = finite_vals[mask]

#     for i, val in tqdm(zip(indices, values), total=len(values), disable=disable_tqdm(), desc=logger.info("Making values unique")):
#         count = seen.get(val, 0)
#         perturbed = np.nextafter(val, -np.inf, dtype=dtype) - count * np.finfo(dtype).eps
#         finite_vals[i] = perturbed
#         seen[val] = count + 1

#     res[is_finite] = finite_vals
#     return res.reshape(arr.shape)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2012 - 2013
# Matías Herranz <matiasherranz@gmail.com>
# Joaquín Tita <joaquintita@gmail.com>
# https://github.com/PyRadar/pyradar
#
# 2020: Erdogan Taskesen: <erdogant@gmail.com> Converted to python3.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import logging

logger = logging.getLogger(__name__)


def median_filter(img, win_size=3):
    """Median Filter.
    
    Median Filter Description
    -------------------------
    Apply a 'median filter' to 'img' with a window size equal to 'win_size'.


    Parameters
    ----------
    img : array-like
        Input image.
    win_size : int, int (default: 3)
        The size of the windows.

    Returns
    -------
    img_filtered : array-like
        Filtered image.

    Examples
    --------
    >>> import findpeaks
    >>> import matplotlib.pyplot as plt
    >>> img = findpeaks.import_example('2dpeaks_image')
    >>> # Resize
    >>> img = findpeaks.stats.resize(img, size=(300,300))
    >>> # Make grey image
    >>> img = findpeaks.stats.togray(img)
    >>> # Scale between [0-255]
    >>> img = findpeaks.stats.scale(img)
    >>> # Filter
    >>> img_filtered = findpeaks.stats.median_filter(img.copy(), win_size=15)
    >>>
    >>> plt.figure()
    >>> fig, axs = plt.subplots(1,2)
    >>> axs[0].imshow(img, cmap='gray'); axs[0].set_title('Input')
    >>> axs[1].imshow(img_filtered, cmap='gray'); axs[1].set_title('Median filter')

    """
    # assert_window_size(win_size)
    if win_size < 3: raise Exception('ERROR: win size must be at least 3')
    if len(img.shape) > 2: raise Exception('ERROR: Image should be 2D. Hint: set the parameter: togray=True')

    N, M = img.shape
    win_offset = int(win_size / 2)
    img_filtered = np.zeros_like(img)

    for i in np.arange(0, N):
        xleft = i - win_offset
        xright = i + win_offset

        if xleft < 0:
            xleft = 0
        if xright >= N:
            xright = N

        for j in np.arange(0, M):
            yup = j - win_offset
            ydown = j + win_offset

            if yup < 0:
                yup = 0
            if ydown >= M:
                ydown = M

            # assert_indices_in_range(N, M, xleft, xright, yup, ydown)

            window = img[xleft:xright, yup:ydown]
            window_median = np.median(window)
            if (window_median is None) or np.isnan(window_median):
                window_median = 0

            img_filtered[i, j] = round(window_median)

    return img_filtered
